{
 "cells": [
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "# Diffie Hellman (DH) Key Exchange\n",
    "\n",
    "![Diffie Hellman Key Exchange Example](resources/012-diffie_hellman_key_exchange.svg)\n",
    "\n",
    "[Diffie-Hellman Key Exchange](https://en.wikipedia.org/wiki/Diffie–Hellman_key_exchange) (DH) is a very widely used method to exchange cryptographic keys. The private keys are not actually exchanged, they are jointly derived. They were invented and published by Whitfield Diffie and Martin Hellman in 1976.\n",
    "\n",
    "In an encrypted communication channel an [symmetric key](https://en.wikipedia.org/wiki/Symmetric-key_algorithm) is used, this means both parties need the same encryption key for either encrypting or decrypting the messages. DH allows to infer this key without the public exchange of it.\n",
    "\n",
    "The algorithm is based on Elliptic Curve Cryptography, a method of doing public-key cryptography based on algebra structure of elliptic curves over finite fields. \n",
    "\n",
    "It is based on the principle that:\n",
    "\n",
    "$$ A = g^{a}\\;mod\\;p $$\n",
    "$$ B = g^{b}\\;mod\\;p $$\n",
    "\n",
    "$$ A^b\\;mod\\;p = g^{ab}\\;mod\\;p = g^{ba}\\;mod\\;p = B^a\\;mod\\;p $$\n",
    "\n",
    "The resulting key for the Diffie Hellman Key Exchange can be then used to any data encryption such as AES, Blowfish etc.\n",
    "\n",
    "## Terminology\n",
    "\n",
    "* private key - $a$ and $b$ secret integer specific for each end between 1 and the prime number $p$. Also called secret.\n",
    "* public key - $A$ and $B$ generated key derived from the private key to be shared\n",
    "* $g$ or $b$ - called generator or base. It is a [primitive root modulo](https://en.wikipedia.org/wiki/Primitive_root_modulo_n) or a small prime number.\n",
    "* $p$ - big [prime number](https://en.wikipedia.org/wiki/Prime_number) often between 2048 and 4096 bits long\n",
    "\n",
    "## Algorithm\n",
    "\n",
    "1. The first party (Alice) picks two numbers, $g$ a [primitive root modulo](https://en.wikipedia.org/wiki/Primitive_root_modulo_n) and $p$ a big [prime number](https://en.wikipedia.org/wiki/Prime_number) and tells them to the second party\n",
    "2. The second party (Bob) picks a secret number $a$, and then it computes $A = g^a\\;mod\\;p$ and sends the result $A$ back to the first party. The secret number $a$ has not been sent only the calculation result.\n",
    "3. Then the first party (Alice) does the same, it selects a secret number $b$ and calculates $B = g^b\\;mod\\;p$ and sends $B$ to the second party (Bob)\n",
    "4. The second party (Bob) takes the received number $B$ and calculates $B^a\\;mod\\;p$\n",
    "5. The first party (Alice) takes the received number $A$ and calculates $A^b\\;mod\\;p$\n",
    "\n",
    "**Both parties will get the same result.**\n",
    "\n",
    "$$ (g^a\\;mod\\;p)^b\\;mod\\;p = g^{ab}\\;mod\\;p $$\n",
    "$$ (g^b\\;mod\\;p)^a\\;mod\\;p = g^{ba}\\;mod\\;p $$"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Imports"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 23,
   "metadata": {},
   "outputs": [],
   "source": [
    "import random"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Algorithm"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 24,
   "metadata": {},
   "outputs": [],
   "source": [
    "class Person:\n",
    "  def __init__(self, seed: int = 1234, name: str = \"\", left: bool = True, verbose: bool = True):\n",
    "    \"\"\"Representing a end of the kexy exchange transmission\n",
    "\n",
    "    Args:\n",
    "        seed (int, optional): seed for initialising the random number generators. Defaults to 1234.\n",
    "        name (string, optional): name of the person. Defaults to \"\".\n",
    "        left (bool, optional): bool indicating if its the left or right party (only used for prints). Defaults to True.\n",
    "        verbose (bool, optional): being verbose or not. Defaults to True.\n",
    "    \"\"\"\n",
    "    self._name = name\n",
    "    self._seed = seed\n",
    "    self._rand_gen = random.Random()\n",
    "    self._rand_gen.seed(self._seed)\n",
    "    self._left = left\n",
    "    self._verbose = verbose\n",
    "    if self._left:\n",
    "      self._prefix = \"\"\n",
    "    else:\n",
    "      self._prefix = \" \"*40\n",
    "    if self._verbose:\n",
    "      print(\"{}{} is alive\".format(self._prefix, self._name))\n",
    "\n",
    "  def genPrimeNumber(self, x: int, y: int):\n",
    "    \"\"\"generates a random prime number between the range x and y, very inefficient method\n",
    "\n",
    "    Args:\n",
    "        x (int): lower range bound\n",
    "        y (int): higher range bound\n",
    "\n",
    "    Returns:\n",
    "        int: random prime number\n",
    "    \"\"\"\n",
    "    # generates list of prime numbers\n",
    "    _prime_list = []\n",
    "    for n in range(x, y):\n",
    "      _isPrime = True\n",
    "      for num in range(2, n):\n",
    "        if n % num == 0:\n",
    "          _isPrime = False\n",
    "\n",
    "      if _isPrime:\n",
    "        _prime_list.append(n)\n",
    "\n",
    "    # select a prime number\n",
    "    return self._rand_gen.choice(_prime_list)\n",
    "\n",
    "  def genSecretKey(self):\n",
    "    \"\"\"generates a random private key between 1 and prime $p$. Representing self._name[0].lower()\"\"\"\n",
    "    self._secret_key = self._rand_gen.randint(1,self.p)\n",
    "    if self._verbose:\n",
    "      print(\"{}{} generated his secret key {}: {}\".format(self._prefix, self._name, self._name[0].lower(), self._secret_key))\n",
    "\n",
    "  def genPublicKey(self):\n",
    "    \"\"\"generates the public variables (key) to be shared. Representing self._name[0].upper()\n",
    "\n",
    "        Returns:\n",
    "        int: shared publiuc key self._name.lower()\n",
    "    \"\"\"\n",
    "    self.public_key = (self.g**self._secret_key) % self.p\n",
    "    if self._verbose:\n",
    "      print(\"{}{} generated his public key {}: {}\".format(self._prefix, self._name, self._name[0].upper(), self.public_key))\n",
    "    return self.public_key\n",
    "\n",
    "  def getPublicKey(self, public_key: int):\n",
    "    \"\"\"get the other party public key\n",
    "\n",
    "    Args:\n",
    "        public_key (int): pubic key from partner\n",
    "    \"\"\"\n",
    "    self.other_public_key = public_key\n",
    "\n",
    "  def genGenerator(self):\n",
    "    \"\"\"generates the public base $b$ or generator number $g$\n",
    "\n",
    "        Returns:\n",
    "        int: shared number $g$\n",
    "    \"\"\"\n",
    "    self.g = self.genPrimeNumber(1,50)\n",
    "    if self._verbose:\n",
    "      print(\"{}{} generated shared base g: {}\".format(self._prefix, self._name, self.g))\n",
    "    return self.g\n",
    "\n",
    "  def setGenerator(self, g: int):\n",
    "    \"\"\"sets the base $b$ or generator $g$ number, if the other party has generated it\n",
    "\n",
    "    Args:\n",
    "        g (int): shared generator number\n",
    "\n",
    "    Returns:\n",
    "        int: shared number $g$\n",
    "    \"\"\"\n",
    "    self.g = g\n",
    "    return self.g\n",
    "\n",
    "  def genPrime(self):\n",
    "    \"\"\"generates a public prime number. This is way too small for reality, commonly 4096 bits\n",
    "\n",
    "    Returns:\n",
    "        int: shared prime number $p$\n",
    "    \"\"\"\n",
    "    self.p = self.genPrimeNumber(100, 1000)\n",
    "    if self._verbose:\n",
    "      print(\"{}{} generated shared prime p: {}\".format(self._prefix, self._name, self.p))\n",
    "    return self.p\n",
    "\n",
    "  def setPrime(self, p: int):\n",
    "    \"\"\"sets the prime number $p$, if the other party has generated it\n",
    "\n",
    "    Args:\n",
    "        p (int): shared prime number $p$\n",
    "\n",
    "    Returns:\n",
    "        int: shared prime number $p$\n",
    "    \"\"\"\n",
    "    self.p = p\n",
    "    return self.p\n",
    "\n",
    "  def genSharedSecret(self):\n",
    "    \"\"\"generates the common shared secret key\n",
    "\n",
    "    Returns:\n",
    "        int: shared secret key\n",
    "    \"\"\"\n",
    "    self._shared_secret = (self.other_public_key ** self._secret_key) % self.p\n",
    "    if self._verbose:\n",
    "      print(\"{}{} generated shared secret: {}\".format(self._prefix, self._name, self._shared_secret))"
   ]
  },
  {
   "cell_type": "markdown",
   "metadata": {},
   "source": [
    "## Test"
   ]
  },
  {
   "cell_type": "code",
   "execution_count": 25,
   "metadata": {},
   "outputs": [
    {
     "name": "stdout",
     "output_type": "stream",
     "text": [
      "Alice is alive\n",
      "                                        Bob is alive\n",
      "Alice generated shared base g: 43\n",
      "Alice generated shared prime p: 257\n",
      "Alice generated his secret key a: 4\n",
      "                                        Bob generated his secret key b: 131\n",
      "Alice generated his public key A: 187\n",
      "                                        Bob generated his public key B: 163\n",
      "Alice generated shared secret: 95\n",
      "                                        Bob generated shared secret: 95\n"
     ]
    }
   ],
   "source": [
    "alice = Person(1234, \"Alice\", True)\n",
    "bob = Person(4321, \"Bob\", False)\n",
    "\n",
    "# generate variables g and p\n",
    "g = alice.genGenerator()\n",
    "p = alice.genPrime()\n",
    "\n",
    "# share variables g and p\n",
    "bob.setGenerator(g)\n",
    "bob.setPrime(p)\n",
    "\n",
    "# create secrets keys\n",
    "alice.genSecretKey()\n",
    "bob.genSecretKey()\n",
    "\n",
    "# create public keys\n",
    "alice_pubkey = alice.genPublicKey()\n",
    "bob_pubkey = bob.genPublicKey()\n",
    "\n",
    "# Exchange keys\n",
    "alice.getPublicKey(bob_pubkey)\n",
    "bob.getPublicKey(alice_pubkey)\n",
    "\n",
    "# Calculate shared secret which should be both times the same\n",
    "alice.genSharedSecret()\n",
    "bob.genSharedSecret()"
   ]
  }
 ],
 "metadata": {
  "interpreter": {
   "hash": "579658931a9ee5a566d859011b110d45b1307b50642b87f1e088320bdb784c2c"
  },
  "kernelspec": {
   "display_name": "Python 3.8.11 64-bit ('daysofalgo': conda)",
   "language": "python",
   "name": "python3"
  },
  "language_info": {
   "codemirror_mode": {
    "name": "ipython",
    "version": 3
   },
   "file_extension": ".py",
   "mimetype": "text/x-python",
   "name": "python",
   "nbconvert_exporter": "python",
   "pygments_lexer": "ipython3",
   "version": "3.8.11"
  },
  "orig_nbformat": 4
 },
 "nbformat": 4,
 "nbformat_minor": 2
}
